//* canvas.h
// celes@tecgraf.puc-rio.br
// Tecgraf/PUC-Rio
// Feb 2003

#ifndef VGL_CANVAS_H
#define VGL_CANVAS_H

#include <ds/list.h>
#include <alg/matrix.h>
#include "viewsurface.h"
#include "notifier.h"
#include "camera.h"

class VglSubCanvas;
class VglHandler;
//class VglCursor;  //TODO

class VGL_API VglCanvas : public VglViewSurface
{
public:
  typedef DsPList<VglSubCanvas> List;
private:
  VglNotifier* m_notifier;
  //VglCursor* m_cursor;
  VglReshapeFunction m_reshape;
  void* m_rdata;
  VglFunction m_redraw;
  void* m_data;
  VglFunction m_eredraw;
  void *m_edata;
  List m_subcanvases;
  bool  m_mapping;
  float m_magfactor;
  float m_dist;
  float m_width, m_height;
  float m_shiftx, m_shifty;
  bool m_redcyan;
  float m_redcyan_rs, m_redcyan_rb, m_redcyan_gs, m_redcyan_gb, m_redcyan_bs, m_redcyan_bb;
  AlgMatrix m_rotation;
protected:
  /**
   * Constructs a new VglCanvas and associates it with a camera and an event notifier.
   * \param camera   Canvas' camera.
   * \param notifier Canvas' event notifier.
   */
  VglCanvas (VglCamera* camera, VglNotifier* notifier);
public:
  /**
   * Destroys a VglCanvas object.
   */
  virtual ~VglCanvas ();
  /**
   * Configures the red & cyan stereo mode.
   */
  void SetRedCyanStereoMode (bool f)
  {
    m_redcyan = f;
  }
  /**
   * Checks if red & cyan stereo mode is enabled.
   */
  bool GetRedCyanStereoMode ()
  {
    return m_redcyan;
  }
  void SetScaleBiasRed (float s, float b)
  {
    m_redcyan_rs = s;
    m_redcyan_rb = b;
  }
  void SetScaleBiasCyan (float gs, float gb, float bs, float bb)
  {
    m_redcyan_gs = gs;
    m_redcyan_gb = gb;
    m_redcyan_bs = bs;
    m_redcyan_bb = bb;
  }
	/**
	 * Sets the event notifier for this canvas.
	 */
	virtual void SetNotifier(VglNotifier* notifier)
	{
		m_notifier = notifier;
	}
  virtual VglNotifier* GetNotifier ()
  {
    return m_notifier;
  }
  /**
   * Adds a new event handler to the canvas.
   * \param handler Event handler.
   * \return Success in the insertion.
   */
  virtual bool AddHandler (VglHandler* handler);
  /**
   * Removes the handler from the canvas.
   * \param handler Event handler.
   * \return Success in the removal.
   */
  virtual bool RemHandler (VglHandler* handler);
  /**
   * Adds a new subcanvas event handler to the canvas.
   * Only to be called internally.
   */
  virtual bool AddSubHandler (VglHandler* handler);
  /**
   * Obtains an iterator to the canvas handler list.
   */
  virtual DsPList<VglHandler>::Iterator GetHandlers () { return m_notifier->GetHandlers(); }
  /**
   * Removes all handlers.
   */
  virtual void ClearHandlers ();
  /**
   * Adds a new subcanvas to the canvas.
   */
  virtual void AddSubCanvas (VglSubCanvas* s);
  /**
   * Removes a subcanvas from the canvas.
   * Returns the success in the removal.
   */
  virtual bool RemSubCanvas (VglSubCanvas* s);
  /**
   * Obtains an iterator to the subcanvas list.
   */
  virtual List::Iterator GetSubCanvases ()
  {
    return m_subcanvases.Begin();
  }
  /**
   * Removes all subcanvases.
   */
  virtual void ClearSubCanvases ();
  /**
   * Adds a new idle function to the canvas.
   * 'data' will be passed to the callback.
   * 'repeat' is the repeat policy (VGL_ONE_TIME or VGL_UNTIL_EVENT or VGL_FOREVER)
   * \return Success in the insertion.
   */
  virtual bool AddIdle (VglFunction func, void* data, int repeat);
  /**
   * Removes an idle function from the canvas.
   * 'data' will be passed to the callback.
   */
  virtual bool RemIdle (VglFunction func, void* data);
  /**
   * Sets the canvas' reshape callback.
   * Callback will be called with the new width, new height and the specified data.
   */
  virtual void SetReshapeFunc (VglReshapeFunction func, void* data=0);
  /**
   * Sets the canvas' redraw callback.
   * 'data' will be passed to the callback.
   */
  virtual void SetRedrawFunc (VglFunction redraw, void* data=0);
  /**
   * Sets the canvas' end redraw callback.
   * 'data' will be passed to the callback.
   */
  virtual void SetEndRedrawFunc (VglFunction eredraw, void* data=0);
	/**
	 * Gets the canvas redraw callback.
	 */
  virtual VglFunction GetRedrawFunc ()
	{
		return m_redraw;
	}
	/**
	 * Gets the canvas redraw callback data.
	 */
  virtual void* GetRedrawData ()
	{
		return m_data;
	}
	/**
	 * Gets the canvas end redraw callback.
	 */
  virtual VglFunction GetEndRedrawFunc ()
	{
		return m_eredraw;
	}
	/**
	 * Gets the canvas end redraw callback data.
	 */
  virtual void* GetEndRedrawData ()
	{
		return m_edata;
	}
	/**
	 * Gets the canvas reshape callback.
	 */
  virtual VglReshapeFunction GetReshapeFunc ()
	{
		return m_reshape;
	}
	/**
	 * Gets the canvas reshape callback data.
	 */
  virtual void* GetReshapeData ()
	{
		return m_rdata;
	}
  /**
   * Loads a camera, updating its parameters in case mapping is used.
   */
  void LoadCamera (VglCamera *camera);
  /**
   * Sets the canvas' keyboard callback.
   * 'data' will be passed to the callback as last parameter.
   */
  virtual void SetKeyboardFunc (VglKeyboardFunction func, void *data=0); 
  /**
   * Sets the canvas' mouse button callback.
   * 'data' will be passed to the callback as last parameter.
   */
  virtual void SetMouseButtonFunc (VglMouseButtonFunction func, void *data=0); 
  /**
   * Sets the canvas' mouse motion callback.
   * 'data' will be passed to the callback as last parameter.
   */
  virtual void SetMouseMotionFunc (VglMouseMotionFunction func, void *data=0);
  /**
   * Sets the canvas' mouse entry callback.
   * 'data' will be passed to the callback as last parameter.
   */
  virtual void SetMouseEntryFunc (VglMouseEntryFunction func, void *data=0);
  /**
   * Activates canvas.
   */
  virtual void Activate () = 0;
  /**
   * Is the canvas active?
   */
  virtual bool IsActive () = 0;
  /**
   * Gets the canvas' width in pixels.
   * \return Canvas width in pixels.
   */
  virtual int GetWidth () = 0;
  /**
   * Gets the canvas' height in pixels.
   * \return Canvas height in pixels.
   */
  virtual int GetHeight () = 0;
  /**
   * Returns if the canvas is a stereo canvas or not.
   */
  virtual bool IsStereo () = 0;
  /**
   * Performs a swap buffer operation in case the view surface has more than one buffer.
   */
  virtual void SwapBuffers () = 0;
  /**
   * Performs a post redraw operation.
   */
  virtual void PostRedraw ();
  /**
   * Performs a redraw operation.
   */
  virtual void Redraw ();
  /**
   * Performs the drawing of an item (addressing issues like camera setup, stereoscopy, etc).
   */
  virtual void DrawItem (VglFunction func, void* data=0);
  virtual void SetCursorPos (float x, float y);
  virtual void SetCamera (VglCamera* camera);
  virtual void SetAudienceDistance (float dist)
  {
    m_dist = dist;
  }
  virtual float GetAudienceDistance () const
  {
    return m_dist;
  }
  virtual void SetDimension (float width, float height)
  {
    m_width = width;
    m_height = height;
  }
  virtual void GetDimension (float *width, float *height) const
  {
    printf("GetDimension not being reimplemented...\n");
    *width = m_width;
    *height = m_height;
  }
  virtual void SetShift (float shiftx, float shifty)
  {
    m_shiftx = shiftx;
    m_shifty = shifty;
  }
  virtual void GetShift (float *shiftx, float *shifty) const
  {
    printf("GetShift not being reimplemented...\n");
    *shiftx = m_shiftx;
    *shifty = m_shifty;
  }
  virtual void SetRotation (float angle, float x=0.0f, float y=1.0f, float z=0.0f)
  {
    m_rotation.Identity();
    m_rotation.Rotate(-angle,x,y,z);
  }
  virtual void SetRotation (const AlgMatrix& r)
  {
    printf("SetRotation not being reimplemented...\n");
    m_rotation = r;
  }
  virtual AlgMatrix GetRotation () const
  {
    return m_rotation;
  }
  virtual void EnableMapping (float magfactor=1.0f);
  virtual void DisableMapping ();
  virtual bool IsMappingEnabled () const
  {
    return m_mapping;
  }
  virtual float GetMappingMagFactor () const
  {
    return m_magfactor;
  }
  //virtual void SetCursor (VglCursor* cursor); //TODO
protected:
  static void ReshapeView (void* data);
  static void RedrawView (void* data);
};

#endif
