//* dcanvasproxy.h
// fabraham@tecgraf.puc-rio.br
// Tecgraf/PUC-Rio
// May 2005

#ifndef VGL_DCANVASPROXY_H
#define VGL_DCANVASPROXY_H

#include <utl/server.h>
#include <ds/array.h>
extern "C" {
#include <lzf/lzf.h>
}
#include <vgl/corba/corbanotifier.h>
#include <errno.h>

template <class CANVASTYPE>
class VGL_API VglDCanvasProxy : public CANVASTYPE
{
  VglCORBANotifier *m_notifier;
  UtlServer *m_server;
  UtlServer::Client m_serverclient;
  DsArray<unsigned char> m_buffer;
  DsArray<unsigned char> m_buffercomp;
  int m_width, m_height;
  DsArray<char*> m_parameters;
  int m_compress_level;
  const char *m_notifier_ior;
  char *m_listenhost;
  int m_listenport;
  char *m_iorhost;
  int m_iorport;
  bool m_reshaping;
  bool m_terminate;
  bool ReceiveString (char *buffer, size_t buffersize)
  {
    int n;
    size_t pos=0;
    for (pos=0; pos<buffersize; pos++)
    {
      n = m_server->Receive(&m_serverclient,buffer+pos,1);
      if (n == -1 || n == 0)
        return false;
      assert(n==1);
      if (buffer[pos] == DCANVAS_SEPARATOR)
      {
        buffer[pos] = '\0';
        return true;
      }
    }
    return false;
  }
  bool SendString (const char *buffer)
  {
    char c = DCANVAS_SEPARATOR;
    return m_server->SendAll(&m_serverclient,buffer,strlen(buffer)) && m_server->SendAll(&m_serverclient,&c,1);
  }
  static bool StringToInt (char *buffer, int *result)
  {
    char *endptr;
    *result = strtol(buffer,&endptr,10);
    return *endptr == '\0';
  }
  DsArray<unsigned char>* Compress ()
  {
#ifdef DEBUG_SERVER
    double t0 = utl_gettime();
#endif
    unsigned long size = m_buffer.Size();
    unsigned long compsize;
#ifdef DEBUG_SERVER
    const char *algorithm = 0;
#endif
    DsArray<unsigned char>* ret;
    switch (m_compress_level)
    {
      case DCANVAS_COMPRESSION_LZF:
        m_buffercomp.Resize(int(1.1*size));
        compsize = lzf_compress(m_buffer.GetArray(),m_buffer.Size(),m_buffercomp.GetArray(),m_buffercomp.Size());
        if (compsize == 0)
        {
          printf("DVGL: An error occurred while compressing the buffer: %s\n",strerror(errno));
          return NULL;
        }
#ifdef DEBUG_SERVER
        algorithm = "liblzf";
#endif
        m_buffercomp.Resize(compsize);
        ret = &m_buffercomp;
        break;
      case DCANVAS_COMPRESSION_NONE:
#ifdef DEBUG_SERVER
        algorithm = "no compression";
#endif
        ret = &m_buffer;
        break;
      default:
        printf("DVGL: Unknown compression algorithm, aborting.\n");
        return NULL;
    }
#ifdef DEBUG_SERVER
    printf(
        "DVGL[send]: Time spent compressing %lu bytes into %lu bytes using %s: %lf ms\n",
        m_buffer.Size(),
        compsize,
        algorithm,
        (utl_gettime()-t0)*1000
        );
#endif
    return ret;
  }
public:
  VglDCanvasProxy () :
    CANVASTYPE(new VglPBufferCanvas(NULL)),
    m_notifier(new VglCORBANotifier()),
    m_server(NULL),
    m_compress_level(DCANVAS_COMPRESSION_NONE),
    m_notifier_ior(NULL),
    m_listenhost(NULL),
    m_listenport(-1),
    m_iorhost(NULL),
    m_iorport(-1),
    m_reshaping(false)
  {
    SetNotifier(m_notifier);
  }
  virtual ~VglDCanvasProxy ()
  {
    if (m_server)
      delete m_server;
    int i;
    for (i=0; i<m_parameters.Size(); i++)
      free(m_parameters[i]);
    delete m_notifier;
  }
  bool SetPort (int port)
  {
    UtlServer *server = new UtlServer(port);
    if (!server->Init())
    {
      delete server;
      return false;
    }
    // success
    if (m_server != NULL)
      delete m_server;
    m_server = server;
    return true;
  }
  void SetNotifierListenHost (const char *listenhost)
  {
    free(m_listenhost);
    m_listenhost = listenhost ? strdup(listenhost) : NULL;
  }
  void SetNotifierListenPort (int listenport)
  {
    m_listenport = listenport;
  }
  void SetNotifierIORHost (const char *iorhost)
  {
    free(m_iorhost);
    m_iorhost = iorhost ? strdup(iorhost) : NULL;
  }
  void SetNotifierIORPort (int iorport)
  {
    m_iorport = iorport;
  }
  void Terminate ()
  {
    VglPBufferCanvas *pbcanvas = (VglPBufferCanvas*)GetCanvas();
    printf("DVGL: Finishing proxy service.\n");
    pbcanvas->Terminate();
    assert(m_server);
    m_server->Disconnect(&m_serverclient);
    delete m_server;
    m_server = NULL;
  }
  int GetParameterNum ()
  {
    return m_parameters.Size();
  }
  const char *GetParameter (int index)
  {
    assert(0 <= index && index < m_parameters.Size());
    return m_parameters[index];
  }
  bool MainLoop ()
  {
    // create corba notifier if possible
    if (m_notifier->Listen(m_listenhost,m_listenport,m_iorhost,m_iorport))
      m_notifier_ior = m_notifier->GetIOR();
    // listen to new connections
    bool ret;
    int w, h, i, n;
    char line[1024];
    VglPBufferCanvas *pbcanvas;
    if (!m_server)
    {
      printf("DVGL: No server info set yet, aborting.\n");
      return false;
    }
    if (!m_server->Listen())
    {
      printf("DVGL: Server was unable to listen to new connections, aborting.\n");
      m_server->ShutDown();
      return false;
    }
    m_terminate = false;
    while (!m_terminate)
    {
      printf("DVGL: Listening for new connections on port %d.\n",m_server->GetPort());
      if (m_server->Accept(&m_serverclient) != UtlServer::ERR_NONE)
        goto error;
      printf("DVGL: New connection from %s.\n",m_serverclient.GetIP());
      // get initial window size
      if (!ReceiveString(line,sizeof(line)) || !StringToInt(line,&m_width) || m_width < 1)
        goto error;
      if (!ReceiveString(line,sizeof(line)) || !StringToInt(line,&m_height) || m_height < 1)
        goto error;
      if (!ReceiveString(line,sizeof(line)) ||
          !StringToInt(line,&m_compress_level) ||
          m_compress_level < 0 ||
          m_compress_level >= DCANVAS_COMPRESSION_MAX)
        goto error;
      if (!ReceiveString(line,sizeof(line)) || !StringToInt(line,&n) || n < 0)
        goto error;
      //printf("Received parameter number: %s\n",line);
      n = atoi(line);
      for (i=0; i<n; i++)
      {
        if (!ReceiveString(line,sizeof(line)))
          goto error;
        m_parameters[i] = strdup(line);
        //printf("Parameter %d = '%s'\n",i,GetParameter(i));
      }
      if (m_notifier_ior)
      {
        if (!SendString("yes") || !SendString(m_notifier_ior))
          goto error;
      }
      else if (!SendString("no"))
        goto error;
      m_parameters.Resize(n);
      pbcanvas = (VglPBufferCanvas*)GetCanvas();
      if (!pbcanvas->Initialize(m_width,m_height,UtlPBuffer::SINGLE))
      {
        printf("DVGL: Failed initializing pbuffer of size %dx%d, aborting.\n",w,h);
        goto error;
      }
      pbcanvas->Activate();
      if (!CANVASTYPE::Init())
      {
          printf("Canvas failed to initialize, goto error\n");
        pbcanvas->Deactivate();
        goto error;
      }
      pbcanvas->MainLoop();
      SetError(true);
      if (!ret)
        goto error;
      goto end;
    error:
      m_terminate = true;
    end:
      m_server->Disconnect(&m_serverclient);
    }
    m_server->ShutDown();
    return true;
  }
  // virtual methods
  virtual bool Init ()
  {
    // do nothing. init will be called on the mainloop
    return true;
  }
  virtual void Redraw ()
  {
    int w = GetWidth();
    int h = GetHeight();
    unsigned int size;
    DsArray<unsigned char>* buffer;
    if (!m_reshaping)
    {
      bool reshaped;
      if (!m_server->ReceiveAll(&m_serverclient,&reshaped,sizeof(reshaped)))
        goto error;
      if (reshaped)
      {
        int reqw, reqh;
        VglPBufferCanvas *pbcanvas = (VglPBufferCanvas*)GetCanvas();
        if (!m_server->ReceiveAll(&m_serverclient,&reqw,sizeof(reqw)) ||
            !m_server->ReceiveAll(&m_serverclient,&reqh,sizeof(reqh)))
          goto error;
        pbcanvas->Reshape(reqw,reqh);
        m_reshaping = true;
        return; // next loop will not receive anything
                // but will redraw with reshaped canvas.
      }
    }
    else
      m_reshaping = false;
    //char line[1024];
#ifdef DEBUG_PROXY
    printf("---------------------------\n");
#endif
    CANVASTYPE::Redraw();
    if (GetError())
      goto error;
    // read framebuffer
    m_buffer.Resize(3*w*h);
    glFlush();
    glPushAttrib(GL_COLOR_BUFFER_BIT | GL_PIXEL_MODE_BIT);
    glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
    if (w%4)
      glPixelStorei(GL_PACK_ALIGNMENT,1);
    else if (w%8)
      glPixelStorei(GL_PACK_ALIGNMENT,4);
    else
      glPixelStorei(GL_PACK_ALIGNMENT,8);
    glReadBuffer(GL_BACK);
    glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,m_buffer.GetArray());
    glPopClientAttrib();
    glPopAttrib();
    // compress
    buffer = Compress();
    if (!buffer)
      goto error;
    size = buffer->Size();
    if (!m_server->SendAll(&m_serverclient,&size,sizeof(size)) ||
        !m_server->SendAll(&m_serverclient,buffer->GetArray(),size))
    {
      printf("DVGL: Client lost connection, aborting.\n");
      goto error;
    }
    return;
  error:
    printf("DVGL: Client lost connection, aborting.\n");
    VglPBufferCanvas *pbcanvas = (VglPBufferCanvas*)GetCanvas();
    pbcanvas->Terminate();
  }
};

#endif
