#ifndef VGL_SELECTION_H
#define VGL_SELECTION_H


#include <vgl/handler.h>
#include <vgl/defines.h>
#include <vgl/hitrecord.h>
#include <vgl/canvas.h>


class VglViewSurface;


/**
 * This class provides the basic functionality for all selection handlers. 
 * It requires the setup of one callback: the selection render callback. 
 * This callback will be called in selection mode and must load/push/pop names 
 * into the name stack, names which will be returned in the selection results.
 *
 * Classes derived from this class must implement the process method, 
 * which receives a list of selection hits (see VglHitRecord for more information).
 *
 * It also defines methods to perform selection:
 * - by picking (clicking).
 * - tracing rays and obtaining the ray intersections with the geometry.
 * - obtaining all visible geometry.
 * - at any modelview/projection configuration.
 */
class VGL_API VglSelectionHandler : public VglHandler
{
public:
  /**
   * Returns the event handler class classname.
   */  
  static char* CLASSNAME () { return "SelectionHandler"; }
  virtual char* ClassName () const { return CLASSNAME(); }
  /**
   * Default constructor.
   * 'surface' is the viewing surface associated with the handler.
   * 'sel_buf_size' is the size of the selection buffer used.
   * 'tolw' and 'tolh' are the picking tolerances in raster coordinates (width and height).
   */
  VglSelectionHandler (VglViewSurface* surface, int sel_buf_size = 512, double tolw = 5.0, double tolh = 5.0);
  /**
   * Destructor.
   */
  virtual ~VglSelectionHandler ();
  /**
   * Selection function data type.
   */  
  typedef void (*SelectionFunc)(void* data);
  /**
   * Sets the rendering function used for selection.
   * 'sel_render_func' is the callback that will be called to render the geometry. It is expected that it pushes/pops names into the name stack.
   * 'data' is the parameter that will be passed to the selection render callback.
   */
  void SetSelectionRenderFunc (SelectionFunc f, void* data);
  /**
   * Hit Process Function (called after the selection rendering function)
   */
  virtual void ProcessFunction (VglSelectionHit* hits, int n) = 0;
  /**
   * Trace a ray and find its intersections with the geometry.
   * This is implemented by setting up a very thin orthographic viewing frustum to simulate the ray. Its dimensions are (2*tolx,2*toly,range).
   * Everything inside this viewing volume will cause a selection hit.
   * 'position' is the ray source
   * 'direction' is the ray direction.
   * 'tolx' and 'toly' are the tolerances in world coordinates.
   * 'range' is the ray range in world coordinates.
   */
  virtual void Trace (AlgVector& position, AlgVector& direction, float tolx = 0.01, float toly = 0.01, float range = 1000.0);
  /**
   * Performs picking at the selected region.
   * 'x' and 'y' are the normalized x and y window coordinates.
   */
  virtual void Pick (float x, float y);
  /**
   * Gets all visible geometry, using the viewing canvas camera information.
   */
  virtual void GetVisible ();
protected:
  bool unproject(VglSelectionHit* h, double winx, double winy, double* worldx, double* worldy, double* worldz);
  bool unproject(double winx, double winy, double winz, double* worldx, double* worldy, double* worldz);
  bool project(double worldx, double worldy, double worldz, double* winx, double* winy, double* winz);
  void PreparePick(float x, float y, double tolw, double tolh);
  void PrepareSelection();
  void FinishSelection();
  void FinishPick();
  /**
   * Performs selection at the current viewing volume configuration.
   */
  void Select();
  
  double m_tolw, m_tolh;
  SelectionFunc m_selectionfunc;
  void* m_selectiondata;
  int m_buffersize;
  unsigned int* m_selectionbuffer;
  VglSelectionHit* m_selectionhit;
  VglViewSurface* m_surface;
};



#endif
