//* videowriter.h
// fabraham@tecgraf.puc-rio.br
// Tecgraf/PUC-Rio
// June 2004

#ifndef VGL_IMAGEWRITER_H
#define VGL_IMAGEWRITER_H

#include <vgl/handler.h>
#include <ds/array.h>
#include <ds/queue.h>
#include <ds/pool.h>
#include <pthread.h>
#include <im.h>

class VglCanvas;

/**
 * Video Writer Handler.
 * This handler will record the frames generated by the application.
 * At the end of recording, it is possible to save the video in the
 * formats supported by IM.
 *
 * The handler will readback what was rendered by the application
 * in a PostRedraw event and will save it into temporary files.
 * 
 * Methods are provided to:
 * - configure the temporary files directory (it is advised to use a
 * directory with the fastest possible writing performance).
 * - configure the maximum application frame rate: controls the frame
 * rate in order to synchronize the video with the application.
 * - save the video.
 */
class VGL_API VglVideoWriter : public VglHandler
{
public:
  /**
   * Video end callback type.
   */  
  typedef void (*VideoEndFunc)(VglVideoWriter *writer, void *data);
private:
  VglCanvas* m_canvas;
  int m_frameindex;
  char* m_tempdir;
  char* m_basefilename;
  char *m_videofilename;
  float m_fps;
  double m_last_t;
  double m_maxfps;
  int m_width, m_height;
  int m_repeat;
  // members for deferred encoding
  VideoEndFunc m_endfunc;
  void *m_enddata;
  typedef struct
  {
    DsArray<unsigned char> *buffer;
    int width;
    int height;
    int index;
  } Buffer;
  DsPool<DsArray<unsigned char> > m_bufferpool;
  DsQueue<Buffer*> m_bufferqueue;
  pthread_mutex_t m_mutexpool;
  pthread_mutex_t m_mutexqueue;
  pthread_cond_t m_condqueuenotempty;
  pthread_cond_t m_condqueuenotfull;
  // members for realtime encoding
  imFile *m_videofile;
  DsArray<unsigned char> m_buffer;
  bool m_encodelater;
  bool m_paused;
  bool m_usemaxfps;
  bool m_recording;
  bool m_handlekeyboard;
  // private methods
  bool BeginRecording ();
  void EndRecording ();
  DsArray<unsigned char>* CreateBuffer (int size);
  void ReleaseBuffer (Buffer *b);
  void EnqueueBuffer (DsArray<unsigned char> *buffer, int w, int h, int index);
  bool WriteBuffer (Buffer *buffer);
public:
  void WriteThread ();
public:
  static char* CLASSNAME () { return "VideoWriter"; }
  virtual char* ClassName () const { return CLASSNAME(); }
  /**
   * Constructor. 'canvas' is the canvas that will have its image recorded on the video.
   */
  VglVideoWriter (VglCanvas* canvas);
  /**
   * Destructor.
   */
  virtual ~VglVideoWriter ();
  /**
   * Specifies if the encoding should be done on the fly or after the recording stops.
   * Encoding on the fly is currently only supported on Windows.
   */
  void SetEncodeLater (bool flag)
  {
    m_encodelater = flag;
  }
  /**
   * Specifies if the keyboard events ('k' starts video, 'K' ends video)
   * should be handled. (default=true)
   */
  void SetHandleKeyboard (bool flag)
  {
    m_handlekeyboard = flag;
  }
  /**
   * Configures if a maximum application frame rate must be specified
   * in order to have the video synchronized with the application.
   * Disabled by default.
   */
  void SetMaxFPS (bool f, double fps=0.0)
  {
    m_usemaxfps = f;
    if (fps > 0.0)
      m_maxfps = fps;
  }
  /**
   * Configures the video writer to repeat the frames a number of times.
   * Useful for static images that must stay the same on the video.
   * Only used in the real-time-encoding method.
   * Default=1.
   */
  void SetRepeat (int r)
  {
    m_repeat = r;
  }
  /**
   * Specifies the video attributes: filename and fps.
   */
  void SetVideoAttributes (char *videofilename, float fps);
  /**
   * Returns the advised frame rate for video saving.
   * In case a maximum frame rate was specified, returns this value.
   */
  double GetSaveFPS ()
  {
    return m_usemaxfps ? m_maxfps : 24.0;
  }
  /**
   * Configures the directory into which the temporary files will be saved. Default=".".
   */
  void SetTemporaryDir (char* directory); 
  /**
   * Configures the base temporary file filename. Default="vgl".
   */
  void SetBaseTempFileName (char* filename);
  /**
   * Starts the video recording.
   */
  virtual bool Record ()
  {
    return BeginRecording();
  }
  /**
   * Pauses/Unpauses the video recording.
   */
  virtual void Pause ()
  {
    m_paused = !m_paused;
  }
  /**
   * Stops the video recording.
   */
  virtual void Stop ()
  {
    EndRecording();
  }
  /**
   * Saves the video with filename 'videofilename' and framerate 'fps.
   * Returns if the save operation was successful.
   */
  bool Save (const char *videofilename, float fps=24.0f);
  /**
   * Configures the video end callback.
   * 'data' will be passed as parameter to the callback.
   */
  void SetVideoEndFunc (VideoEndFunc func, void *data)
  {
    m_endfunc = func;
    m_enddata = data;
  }
  virtual int Keyboard (int k, int st, float x, float y);
  virtual int PostRedraw ();
};

#endif
