//* notifier.h
// celes@tecgraf.puc-rio.br
// Tecgraf/PUC-Rio
// Jan 2003

#ifndef VGL_NOTIFIER_H
#define VGL_NOTIFIER_H

#include "defines.h"
#include "ds/list.h"

class VglCanvas;
class VglHandler;

class VGL_API VglNotifier
{
protected:
  /**
   * Function that will be called whenever an idle function is added.\n
   * Useful to setup an idle function only when needed (when there is
   * at least 1 idle function)
   */
  virtual void IdleAddRef () = 0;
  /**
   * Function that will be called whenever an idle function is removed.\n
   * Useful to unset an idle function if there is no idle function left.\n
   */
  virtual void IdleRemRef () = 0;
public:
  typedef DsPList<VglNotifier> NotifierList;
  typedef DsPList<VglHandler> HandlerList;
  /**
   * Constructs a new event notifier.
   * \param source Notifier identifier.
   */
  VglNotifier (int source);
  /**
   * Destroys a VglNotifier object.
   */
  virtual ~VglNotifier ();
  /**
   * Activates the event notifier.
   */
  void Activate ();
  /**
   * Deactivates the event notifier.
   */
  void Deactivate ();
  /**
   * Is the notifier active?
   * \return Notifier active = true, false otherwise.
   */
  bool IsActive ();
  /**
   * Registers a new event handler.
   * if 'append', added to the end of the list, else add to the beggining.
   * Returns the success in registration (duplication is not allowed).
   */
  bool AddHandler (VglHandler* handler, bool append=true);
  /**
   * Removes an event handler.
   * \param handler Event handler.
   * \return Success in removal.
   */
  bool RemHandler (VglHandler* handler);
  /**
   * Gets the notifier's event handlers.
   * \return An iterator to the event handler list.
   */
  HandlerList::Iterator GetHandlers ()
  {
    return m_handlers.Begin();
  }
  /**
   * Removes all handlers.
   */
  void ClearHandlers ();
  /**
   * Inserts a new idle function.
   * \param func Idle function.
   * \param data Parameter to be passed to this idle function.
   * \param repeat Repeat policy. (VGL_ONE_TIME / VGL_UNTIL_EVENT / VGL_FOREVER)
   * \return Success in the insertion.
   */
  virtual bool AddIdle (VglFunction func, void* data, int repeat);
  /**
   * Removes an idle function.
   * \param func Idle function.
   * \param data Parameter to the idle function. Must be the same as in the insertion.
   * \return Success in the removal.
   */
  virtual bool RemIdle (VglFunction func, void* data);
  /**
   * Calls the idle functions.
   */
  void CallIdleFuncs ();
  /**
   * Cleans the idle function list.
   */
  void CleanIdleList ();
  /**
   * Resets the idle function list.
   */
  void ResetIdleList ();
	/**
	 * Sets all keys as unpressed.
	 */
	void ResetKeys ();

  /**
   * Handles a keyboard event. Will call all event handlers and the keyboard callback.
   * \param k  Key code.
   * \param st Key state (VGL_PRESS / VGL_RELEASE)
   * \param x  Normalized x coordinate.
   * \param y  Normalized y coordinate.
   * \param c  Control key state (1=pressed,0=unpressed)
   * \param s  Shift key state (1=pressed,0=unpressed)
   * Returns if this event was filtered by any handler.
   */
  bool HandleKeyboard (int k, int st, float x, float y, int c, int s);
  /**
   * Handles a mouse click event. Will call all event handlers and the mouse click callback.
   * \param bt Button (VGL_BUTTON_LEFT / VGL_BUTTON_MIDDLE / VGL_BUTTON_RIGHT)
   * \param st Button state (VGL_PRESS / VGL_RELEASE)
   * \param x  Normalized x coordinate.
   * \param y  Normalized y coordinate.
   * \param c  Control key state (1=pressed,0=unpressed)
   * \param s  Shift key state (1=pressed,0=unpressed)
   * Returns if this event was filtered by any handler.
   */
  bool HandleMouseButton (int bt, int st, float x, float y, int c, int s);
  /**
   * Handles a mouse motion event. Will call all event handlers and the mouse motion callback.
   * \param x  New normalized x coordinate.
   * \param y  New normalized y coordinate.
   * Returns if this event was filtered by any handler.
   */
  bool HandleMouseMotion (float x, float y); 
  /**
   * Handles a mouse entry event. Will call all event handlers and the mouse entry callback.
   * \param st true if mouse entered, false if mouse left.
   * Returns if this event was filtered by any handler.
   */
  bool HandleMouseEntry (bool st); 
  /**
   * Will call PreRedraw() for all event handlers before calling the redraw callback of the canvas.
   * Returns if this event was filtered by any handler.
   */
  bool HandlePreRedraw (void); 
  /**
   * Will call PostRedraw() for all event handlers after calling the redraw callback of the canvas.
   * Returns if this event was filtered by any handler.
   */
  bool HandlePostRedraw (void); 
  /**
   * Is key with code 'k' pressed?
   */
  bool GetKey (int k)
  {
    return m_key[k];
  }
  /**
   * Is control key pressed?
   * \return true if pressed, false otherwise.
   */
  bool GetCtrl ()
  {
    return m_ctrl;
  }
  /**
   * Is shift key pressed?
   * \return true if pressed, false otherwise.
   */
  bool GetShift ()
  {
     return m_shift;
  }
  /**
   * Returns the x coordinate where the button was pressed.
   * \return Normalized x coordinate.
   */
  float GetX0 ()
  {
    return m_x0;
  }
  /**
   * Returns the y coordinate where the button was pressed.
   * \return Normalized y coordinate.
   */
  float GetY0 ()
  {
    return m_y0;
  }
  /**
   * Returns the last x coordinate the mouse visited.
   * \return Normalized x coordinate.
   */
  float GetLastX ()
  {
    return m_lastx;
  }
  /**
   * Returns the last y coordinate the mouse visited.
   * \return Normalized y coordinate.
   */
  float GetLastY ()
  {
    return m_lasty;
  }
  /**
   * Sets the keyboard callback.
   * \param func Keyboard callback.
   * \param data Data that will be passed to this callback.
   */
  void SetKeyboardFunc 
    (void (*func)(int k, int st, float x, float y, void* data), void* data);
  /**
   * Sets the mouse button callback.
   * \param func Mouse button callback.
   * \param data Data that will be passed to this callback.
   */
  void SetMouseButtonFunc 
    (void (*func) (int bt, int st, float x, float y, void* data), void* data);
  /**
   * Sets the mouse motion callback.
   * \param func Mouse motion callback.
   * \param data Data that will be passed to this callback.
   */
  void SetMouseMotionFunc
    (void (*func) (int bt, float x, float y, void* data), void* data);
  /**
   * Sets the mouse entry callback.
   * \param func Mouse entry callback.
   * \param data Data that will be passed to this callback.
   */
  void SetMouseEntryFunc
    (void (*func) (int st, void* data), void* data);
  /**
   * Notify a new keyboard event.
   * \param source Event source.
   * \param k      Key code.
   * \param st     Key state (VGL_PRESS / VGL_RELEASE)
   * \param x      Normalized x coordinate.
   * \param y      Normalized y coordinate.
   * \param c      Control key state (1=pressed,0=unpressed)
   * \param s      Shift key state (1=pressed,0=unpressed)
   */
  static void NotifyKeyboard 
    (int source, int k, int st, float x, float y, int ctl, int sht);
  /**
   * Notifies a new mouse click event.
   * \param source Event source.
   * \param bt     Button (VGL_BUTTON_LEFT / VGL_BUTTON_MIDDLE / VGL_BUTTON_RIGHT)
   * \param st     Button state (VGL_PRESS / VGL_RELEASE)
   * \param x      Normalized x coordinate.
   * \param y      Normalized y coordinate.
   * \param c      Control key state (1=pressed,0=unpressed)
   * \param s      Shift key state (1=pressed,0=unpressed)
   */
  static void NotifyMouseButton 
    (int source, int bt, int st, float x, float y, int ctl, int sht);
  /**
   * Notifies a new mouse motion event.
   * \param source Event source.
   * \param x      New normalized x coordinate.
   * \param y      New normalized y coordinate.
   */
  static void NotifyMouseMotion (int source, float x, float y);
  /**
   * Notifies a new mouse entry event.
   * \param source Event source.
   * \param st     true if mouse entered, false if mouse left.
   */
  static void NotifyMouseEntry (int source, bool status); 
  /**
   * Calls all notifiers idle functions.
   */
  static void DoIdle ();
private:
  void SetLastXY (float x, float y);
  static NotifierList s_notifiers;

  struct Idle
  {
    VglFunction func;
    void* data;
    int   repeat;
    Idle () : func(0), data(0), repeat(0) { }
    Idle (VglFunction f, void* d, int r) : func(f), data(d), repeat(r) { }
    bool operator== (const Idle& oth) 
    { 
      return func==oth.func && 
        data==oth.data && 
        repeat==oth.repeat; 
    }
  };
  typedef DsList<Idle> IdleList;

  int m_source;
  bool m_ctrl, m_shift;
  bool m_key[VGL_MAX_KEY];
  int m_button;
  float m_x0, m_y0;
  float m_lastx, m_lasty;
  float m_setlastx, m_setlasty;
  bool m_setlastxy;
  HandlerList m_handlers;
  IdleList m_idlelist;

  void (*m_keyboardfunc)(int k, int st, float x, float y, void* data);
  void (*m_buttonfunc) (int bt, int st, float x, float y, void* data);
  void (*m_motionfunc) (int bt, float x, float y, void* data);
  void (*m_entryfunc) (int st, void* data);
  void* m_keyboardfuncdata;
  void* m_buttonfuncdata;
  void* m_motionfuncdata;
  void* m_entryfuncdata;
  friend class VglCanvas;
  friend class VglSubCanvas;
};

#endif
