//* subcanvas.h
// fabraham@tecgraf.puc-rio.br
// Tecgraf/PUC-Rio
// Mar 2005

#ifndef VGL_SUBCANVAS_H
#define VGL_SUBCANVAS_H

#include "canvas.h"
#include "viewsurface.h"
#include "camera.h"

class VglSubHandler;
class VglSubNotifier;

class VGL_API VglSubCanvas : public VglViewSurface
{
  VglReshapeFunction m_reshape;
  void* m_rdata;
  VglFunction m_redraw;
  void* m_data;
  VglCanvas *m_parent;
  float m_x, m_y, m_w, m_h, m_winv, m_hinv;
  unsigned char m_cc[4];
  float m_cd;
  VglSubNotifier *m_notifier;
  VglSubHandler *m_handler;
  bool m_enabled;
public:
  /**
   * Constructs a new VglSubCanvas and associates it with a camera.
   * Its parent must also be supplied.
   */
  VglSubCanvas (VglCamera* camera, VglCanvas *parent);
  /**
   * Destroys a VglSubCanvas object.
   */
  virtual ~VglSubCanvas ();
  VglSubHandler *GetSubHandler ();
  /**
   * Sets the subcanvas' window. 'x', 'y', 'w' and 'h' are relative to
   * the subcanvas' parent canvas window width and height and must range
   * from 0.0 to 1.0.
   */
  void SetWindow (float x, float y, float w, float h);
  /**
   * Obtains the subcanvas' window.
   */
  void GetWindow (float *x, float *y, float *w, float *h)
  {
    *x = m_x;
    *y = m_y;
    *w = m_w;
    *h = m_h;
  }
  /**
   * Specifies if the subcanvas handler must filter events to subsequent
   * handlers. [default=true]
   */
  void SetFilterEvents (bool f);
  /**
   * Obtains the subcanvas' x coord in pixels
   */
  int GetX ()
  {
    return int(m_x*m_parent->GetWidth());
  }
  /**
   * Obtains the subcanvas' y coord in pixels
   */
  int GetY ()
  {
    return int(m_y*m_parent->GetHeight());
  }
  /**
   * Obtains the subcanvas' width in pixels
   */
  int GetWidth ()
  {
    return int(m_w*m_parent->GetWidth());
  }
  /**
   * Obtains the subcanvas' height in pixels
   */
  int GetHeight ()
  {
    return int(m_h*m_parent->GetHeight());
  }
  /**
   * Enables/disables a subcanvas.
   */
  void SetEnabled (bool f);
  /**
   * Is subcanvas enabled?
   */
  bool IsEnabled ()
  {
    return m_enabled;
  }
  /**
   * Adds a new event handler to the subcanvas.
   * Returns the success in the insertion.
   */
  bool AddHandler (VglHandler* handler);
  /**
   * Removes the handler from the canvas.
   * Returns the success in the removal.
   */
  bool RemHandler (VglHandler* handler);
  /**
   * Obtains an iterator to the subcanvas handler list.
   */
  VglNotifier::HandlerList::Iterator GetHandlers ();
  /**
   * Removes all handlers.
   */
  void ClearHandlers ();
  /**
   * Redraws the subcanvas.
   */
  void RedrawView ();
  /**
   * Specifies the subcanvas clear color.
   */
  void ClearColor (float r, float g, float b, float a)
  {
    m_cc[0] = (unsigned char)(255*r);
    m_cc[1] = (unsigned char)(255*g);
    m_cc[2] = (unsigned char)(255*b);
    m_cc[3] = (unsigned char)(255*a);
  }
  /**
   * Specifies the subcanvas clear depth value.
   */
  void ClearDepth (float d)
  {
    m_cd = d;
  }
  /**
   * Clears the desired buffers (similar to glClear)
   */
  void Clear (unsigned int mask);
  /**
   * Sets the subcanvas' camera.
   */
  virtual void SetCamera (VglCamera *camera);
  /**
   * Sets the subcanvas' reshape callback.
   * Callback will be called with the new width, new height and the specified data.
   */
  virtual void SetReshapeFunc (VglReshapeFunction func, void* data=0);
  /**
   * Sets the subcanvas' redraw callback.
   * 'data' will be passed to the callback.
   */
  virtual void SetRedrawFunc (VglFunction func, void* data=0);
  /**
   * Gets the subcanvas redraw callback.
   */
  virtual VglFunction GetRedrawFunc ();
  /**
   * Gets the subcanvas redraw callback data.
   */
  virtual void* GetRedrawData ();
  /**
   * Gets the subcanvas reshape callback.
   */
  virtual VglReshapeFunction GetReshapeFunc ();
  /**
   * Gets the subcanvas reshape callback data.
   */
  virtual void* GetReshapeData ();
  /**
   * Requests the redraw of this subcanvas' parent.
   */
  virtual void Redraw ();
  /**
   * Requests the redraw of this subcanvas' parent when the process becomes idle.
   */
  virtual void PostRedraw ();
  virtual void DrawItem (VglFunction func, void* data=0);
  virtual void SetCursorPos (float x, float y);
  virtual void SetRasterCursor (bool f);
};

#endif
